import analysis
import array

inf = float('inf')
class SpikeTimesData:
    def __init__(self):
        self.spike_times = []
        self.types = []
        self._time_points = {}
        self._last_t = -1
        self._cell_groups = {}


    def time(self, t):
        """Return the time denoted by t.
        
        Note:
        
        t may be a numeric value, or a named time point.
        """
        if type(t) in (int, long, float):
            return t
        if t in self._time_points:
            return self._time_points[t]
        raise Exception('Unknown time point: ' + t)


    def _forget_after(self, forget_t):
        for i, ts in enumerate(self.spike_times):
            stop_j = 0
            for j, t in enumerate(ts):
                if t < forget_t:
                    stop_j = j
                else:
                    break
            self.spike_times[i] = ts[: stop_j]
        self._last_t = forget_t



    def _update(self, sim):
        self._cell_groups = sim._cell_groups
        self.types = sim.types
        for name, val in zip(sim._time_points.keys(), sim._time_points.values()):
            self._time_points[name] = val
        
        if not self.spike_times:
            self.spike_times = sim.spike_times
        else:
            mint = self._last_t
            for st, simst in zip(self.spike_times, sim.spike_times):
                st.extend([t for t in simst if t > mint])            
            
        self._last_t = sim.t
        self.run_time = sim.t
        
    def _get_spike_times(self):
        """Return a list of lists of spike times for each cell."""
        return self.spike_times

    def cv(self, cells=None, cell_type=None, time_range=(-inf, inf), combine_cells=False):
        """Return the CVs of cells.
        
        Deprecated Parameters:
            cell_type
        
        Note:
        
        Returns the CVs of all cell types if cells = None.
        
        Results limited to those spikes occuring in the interval time_range.
        
        If combine_cells is False, then returns a list of CVs, one for each
        neuron. If true, returns the average cv of all neurons of cell_type.
        """
        isis = self.isi(cells, cell_type, time_range, False)
        cvs = [analysis.cv(i) for i in isis]
        if combine_cells:
            cvs = sum(cvs) / float(len(cvs))
        return cvs        

    def isi(self, cells=None, cell_type=None, time_range=(-inf, inf), combine_cells=False):
        """Return the Interspike intervals (ISIs) of cells.
        
        Deprecated Parameters:
            cell_type
        
        Note:
        
        Returns the ISIs of all cell types if cells = None.
        
        Results limited to those spikes occuring in the interval time_range.
        
        If combine_cells is False, then returns a list of lists of ISIs, one
        list for each neuron. If True, returns a single list of ISIs covering
        all neurons of the cell_type.
        """
        names = self.types
        spikes = self._get_spike_times()
        isis = []
        time_range = [self.time(time_range[0]), self.time(time_range[1])]
        t_min = min(time_range)
        t_max = max(time_range)
        if cells is None and cell_type is not None:
            cells = cell_type
            warn(DeprecationWarning('cell_type is deprecated; use cells instead.'))
        for i in self.get_indices(cells):
            isi_current = []
            current_spikes = spikes[i]
            for j in xrange(1, len(current_spikes)):
                if current_spikes[j - 1] >= t_min and current_spikes[j] <= t_max:
                    isi_current.append(current_spikes[j] - current_spikes[j - 1])
            isis.append(array.array('d', isi_current))
        if combine_cells == True:
            isi_all = []
            for isi in isis:
                isi_all += isi
            isis = isi_all
        return isis

    def get_indices(self, cells = None):
        """Return the indices of the cells specified.
        
        Parameter:
            cells -- a list of indices or a string naming a cell_type or cell_group
        
        Note:
        
        If cells is None, returns all indices.
        """
        if cells is None:
            return range(len(self.types))
        if isinstance(cells, list):
            return cells
        if cells in self._cell_groups:
            return self._cell_groups[cells]
        return [i for i in range(len(self.types)) if self.types[i] == cells]


    def time(self, t):
        """Return the time corresponding to t.
        
        Note:
        
        t may either be a numeric value or the name of a saved time point.
        """
        if type(t) in (int, long, float):
            return t
        if t in self._time_points:
            return self._time_points[t]
        print 'time_points: ', self._time_points
        raise Exception('Unknown time point: ' + t)

